from pydantic import BaseModel
from typing import Callable, Any, Union
from typing_extensions import Self


class Dependency:
    def __init__(self: Self, func: Callable[...,]) -> None:
        self.func = func


class DependencySchema(BaseModel):
    func: Callable[..., Any]
    deps: list["DependencySchema"]

    @classmethod
    def from_func(
        cls: type[Self],
        func: Callable[..., Any],
        visited: Union[Callable[..., Any], None],
        ancestors: Union[Callable[..., Any], None],
    ) -> Self:
        if visited is None:
            visited = []
        if ancestors is None:
            ancestors = []

        deps: list[Self] = _resolve_deps(func, visited, ancestors)

        return cls(func=func, deps=deps)


def _resolve_deps(
    func: Callable[..., Any],
    visited: list[Callable[..., Any]],
    ancestors: list[Callable[..., Any]],
) -> list[DependencySchema]:
    ...



import pytest

from aws_spy import (
    CORS,
    Authorizer,
    HTTPApi,
    Provider,
    ServerlessConfig,
    SpyAPI,
    build_cognito_issue_url,
)


@pytest.fixture(scope="session")
def authorizer() -> Authorizer:
    return Authorizer(
        issuerUrl=build_cognito_issue_url("eu-central-1_abc123"),
        audience=["abc123"],
    )


@pytest.fixture(scope="session")
def cors() -> CORS:
    return CORS(
        allowedHeaders=["Content-Type", "Authorization"],
        exposedResponseHeaders=["X-Total-Count"],
        allowedMethods=["GET", "POST", "OPTIONS", "PUT", "PATCH", "DELETE"],
        allowedOrigins=["*"],
    )


@pytest.fixture(scope="session")
def http_api(authorizer: Authorizer, cors: CORS) -> HTTPApi:
    return HTTPApi(authorizers={"jwt": authorizer}, cors=cors)


@pytest.fixture(scope="session")
def provider(http_api: HTTPApi) -> Provider:
    return Provider(
        region="us-east-1",
        role="arn:aws:iam::123456789:role/TestLambdaRole",
        httpApi=http_api,
    )


@pytest.fixture(scope="session")
def config(provider: Provider) -> ServerlessConfig:
    return ServerlessConfig(
        service="lambdas",
        plugins=["serverless-python-requirements"],
        provider=provider,
    )


@pytest.fixture
def app(config: ServerlessConfig) -> SpyAPI:
    return SpyAPI(config=config)



import json
from typing import Any, Literal, TypeVar, Union

from pydantic import BaseModel, ValidationError

from aws_spy.core.schemas_utils import ParamSchema

RequestBodyType = TypeVar("RequestBodyType", bound=BaseModel)


def export_params_from_event(
    in_event_params: Union[dict[str, Any], None],
    expected_params: list[ParamSchema],
    type_: Literal["path", "header", "query"],
) -> tuple[dict[str, Any], list[str]]:
    if in_event_params is None:
        in_event_params = {}
    args, errors = {}, []
    for expected_param in expected_params:
        param = in_event_params.get(expected_param.name)
        if param is None and expected_param.is_required:
            errors.append(
                f"Required parameter {expected_param.name} not found in {type_}."
            )
            continue
        try:
            param = expected_param.annotation(param) if param is not None else None
        except ValueError:
            errors.append(
                f"{expected_param.name} should be "
                f"{expected_param.annotation.__name__} type."
            )
            continue
        args[expected_param.arg_name] = param

    return args, errors


def export_request_body(
    body: str, request_body_class: type[RequestBodyType]
) -> tuple[Union[RequestBodyType, dict[str, Any], None], list[str]]:
    try:
        body = json.loads(body)
    except json.JSONDecodeError:
        return None, ["Request body is empty!"]
    try:
        request_body = request_body_class.parse_obj(body)
    except ValidationError as e:
        errors = []
        for error in e.errors():
            if error["type"].startswith("type_error"):
                errors.append(
                    f'Wrong type received at: {error["loc"][0]}. '
                    f'Expected: {error["type"].split(".")[-1]}'
                )
                continue
            if error["type"].endswith("missing"):
                errors.append(f'Value not found at: {error["loc"][0]}')
                continue
            errors.append(f"Unknown error: {error}")  # pragma: no cover
        return None, errors

    return request_body, []



from abc import ABC, abstractmethod
from typing import Any, Union

from pydantic import BaseModel
from typing_extensions import Self  # type: ignore


class BaseResponseSPY(ABC):
    route: Union[BaseModel, None]

    @property
    @abstractmethod
    def response(self: Self) -> dict[str, Any]:
        raise NotImplementedError  # pragma: no cover



from abc import ABC
from enum import Enum
from typing import Union

from typing_extensions import Self  # type: ignore


class ParamType(str, Enum):
    QUERY = "query"
    PATH = "path"
    HEADER = "header"


class Param(ABC):
    in_: ParamType

    def __init__(self: Self, name: Union[str, None]) -> None:
        self.name = name


class QueryClass(Param):
    in_ = ParamType.QUERY


class PathClass(Param):
    in_ = ParamType.PATH


class HeaderClass(Param):
    in_ = ParamType.HEADER



import json
from enum import Enum
from typing import Any, Union

from openapi_schema_pydantic.v3.v3_0_3 import OpenAPI
from openapi_schema_pydantic.v3.v3_0_3.parameter import Parameter
from openapi_schema_pydantic.v3.v3_0_3.util import (
    PydanticSchema,
    Schema,
    construct_open_api_with_schema_class,
)

from aws_spy.core.schemas import Methods, SpyRoute
from aws_spy.core.types import is_type_required

SCHEMA_PARAM = {
    int: "integer",
    str: "string",
    bool: "boolean",
    float: "number",
    Enum: "string",
}


def _get_path_item(method: Methods, route: SpyRoute) -> dict[str, Any]:
    item = {
        "responses": {
            route.status_code: {
                "description": "Successful response",
                "content": {
                    "application/json": {
                        "schema": PydanticSchema(schema_class=route.response_class)
                        if route.response_class
                        else {}
                    }
                },
            },
            422: {
                "description": "Validation Error",
                "content": {
                    "application/json": {
                        "schema": {"$ref": "#/components/schemas/SPY422ErrorResponse"}
                    }
                },
            },
        }
    }
    item["parameters"] = [  # type: ignore
        Parameter(
            name=route_param.name,
            required=is_type_required(route_param.annotation),
            param_in=route_param.in_.in_,
            param_schema=Schema(
                type=SCHEMA_PARAM.get(route_param.annotation, "string"),
                enum=route_param.enum,
            ),
        )
        for route_param in route.header_params + route.path_params + route.query_params
    ]
    if route.tags:
        item["tags"] = route.tags  # type: ignore
    if route.summary:
        item["summary"] = route.summary  # type: ignore
    if route.description:
        item["description"] = route.description  # type: ignore
    if method not in (Methods.GET, Methods.DELETE):
        item.update(
            {
                "requestBody": {
                    "content": {  # type: ignore
                        "application/json": {
                            "schema": PydanticSchema(schema_class=route.request_body)
                            if route.request_body
                            else {}
                        }
                    }
                }
            }
        )

    return item


def construct_base_open_api(
    title: str, version: Union[str, None], routes: dict[str, dict[Methods, SpyRoute]]
) -> OpenAPI:
    info = {"title": title}
    if version is not None:
        info["version"] = version
    return OpenAPI.parse_obj(
        {
            "info": info,
            "paths": {
                path: {
                    method: _get_path_item(method, route)
                    for method, route in route_dict.items()
                }
                for path, route_dict in routes.items()
            },
            "components": {
                "schemas": {
                    "SPY422SingleError": {
                        "title": "SPY 422 Single Error Schema",
                        "required": ["message"],
                        "type": "object",
                        "properties": {
                            "message": {
                                "title": "Error message",
                                "type": "string",
                            }
                        },
                    },
                    "SPY422ErrorResponse": {
                        "title": "SPY 422 Error Response",
                        "required": ["error"],
                        "type": "object",
                        "properties": {
                            "errors": {
                                "type": "array",
                                "items": {
                                    "$ref": "#/components/schemas/SPY422SingleError"
                                },
                                "example": [
                                    "Wrong type received at: age. Expected: integer",
                                    "Value not found at: name",
                                    "Required parameter authorization not found in header.",
                                    "user_id should be UUID type.",
                                ],
                            }
                        },
                    },
                }
            },
        }
    )


def get_openapi(
    title: str, version: Union[str, None], routes: dict[str, dict[Methods, SpyRoute]]
) -> dict[str, Any]:
    return json.loads(
        construct_open_api_with_schema_class(
            construct_base_open_api(title, version, routes)
        ).json(by_alias=True, exclude_none=True, indent=2)
    )



from typing import Any, Union

from aws_spy.core import params


def Query(name: Union[str, None] = None):
    return params.QueryClass(name)


def Path(name: Union[str, None] = None):
    return params.PathClass(name)


def Header(name: Union[str, None] = None):
    return params.HeaderClass(name)



import pytest
from pydantic import BaseModel

from aws_spy import SpyAPI
from aws_spy.core.exceptions import RouteDefinitionException
from aws_spy.core.schemas import Methods


class Request(BaseModel):
    ...


@pytest.mark.parametrize(
    "method",
    [method for method in Methods if method in (Methods.GET, Methods.DELETE)],
)
def test_get_method_with_request_body(app: SpyAPI, method: Methods) -> None:
    app_method = getattr(app, method.value)
    with pytest.raises(
        RouteDefinitionException,
        match=f'{method.upper()} method on "/some_path" cannot have request body!',
    ):

        @app_method("/some_path", "test-lambda")
        def handler(request: Request) -> None:
            ...


@pytest.mark.parametrize(
    "method",
    [method for method in Methods if method not in (Methods.GET, Methods.DELETE)],
)
def test_request_body(app: SpyAPI, method: Methods) -> None:
    app_method = getattr(app, method.value)
    path = "/somepath"

    @app_method(path, "lambda")
    def handler(request: Request) -> None:
        ...

    assert len(app.routes) == 1
    route = app.routes[path][method]
    assert route.request_body == Request
    assert route.request_body_arg_name == "request"



from functools import wraps
from typing import Any, Union

from pydantic import BaseModel
from typing_extensions import Self  # type: ignore

from aws_spy.core.event_utils import export_params_from_event, export_request_body
from aws_spy.core.exceptions import RouteDefinitionException
from aws_spy.core.responses import BaseResponseSPY
from aws_spy.core.schemas import LH, Decorator, Methods, ServerlessConfig, SpyRoute
from aws_spy.responses import ErrorResponse, JSONResponse


class _SPY:
    routes: dict[str, dict[Methods, SpyRoute]]
    function_unique_ids: set[str]

    def __init__(self: Self, prefix: Union[str, None] = None) -> None:
        self.routes = {}
        self.function_unique_ids = set()
        if isinstance(prefix, str) and not prefix.startswith("/"):
            prefix = "/" + prefix
        self.prefix = prefix or ""

    def register_router(self: Self, router: Self) -> None:
        for path, methods in router.routes.items():
            for method, route in methods.items():
                self.add_route(path, method, route)

    def add_route(self: Self, path: str, method: Methods, route: SpyRoute) -> None:
        if not path.startswith("/"):
            path = "/" + path
        path = self.prefix + path
        if route.name in self.function_unique_ids:
            raise RouteDefinitionException(
                f"There is already {route.name} lambda registered."
            )

        if path in self.routes.keys() and method in self.routes[path]:
            raise RouteDefinitionException(
                f'There is already existing "{method.upper()}" '
                'method definition under "{path}" path.'
            )

        if path not in self.routes.keys():
            self.routes[path] = {}

        self.function_unique_ids.add(route.name)
        self.routes[path][method] = route

    def route(
        self: Self,
        *,
        method: Methods,
        path: str,
        name: str,
        authorizer: Union[str, None] = None,
        response_class: Union[type[BaseModel], None] = None,
        status_code: Union[int, None] = None,
        tags: Union[list[str], None] = None,
        summary: Union[str, None] = None,
        description: Union[str, None] = None,
        use_vpc: Union[bool, None] = None,
        skip_validation: Union[bool, None] = None,
    ) -> Decorator:
        def decorator(handler: LH) -> LH:
            route = SpyRoute(
                method=method,
                path=path,
                name=name,
                handler=handler,
                response_class=response_class,
                status_code=status_code,
                tags=tags,
                summary=summary,
                description=description,
                authorizer=authorizer,
                use_vpc=use_vpc,
                skip_validation=skip_validation,
            )
            self.add_route(path, method, route)

            @wraps(handler)
            def wrapper(*args) -> dict[str, Any]:
                if route.skip_validation:
                    return handler(*args)
                event = args[0]
                context = args[1]
                kwargs, errors = {}, []
                for params, errors_ in [
                    export_params_from_event(
                        event.get("pathParameters"), route.path_params, "path"
                    ),
                    export_params_from_event(
                        event.get("headers"), route.header_params, "header"
                    ),
                    export_params_from_event(
                        event.get("queryStringParameters"), route.query_params, "query"
                    ),
                ]:
                    kwargs.update(params)
                    errors += errors_

                if route.request_body and route.request_body_arg_name:
                    request_body, request_body_errors = export_request_body(
                        event.get("body", ""), route.request_body
                    )
                    errors += request_body_errors
                    kwargs[route.request_body_arg_name] = request_body

                if errors:
                    return ErrorResponse(errors, status_code=422).response
                if route.add_event:
                    kwargs["event"] = event
                if route.add_context:
                    kwargs["context"] = context

                return_obj = handler(**kwargs)
                if not isinstance(return_obj, BaseResponseSPY):
                    return_obj = JSONResponse(return_obj)
                return_obj.route = route

                return return_obj.response

            return wrapper

        return decorator

    def get(
        self: Self,
        path: str,
        name: str,
        *,
        authorizer: Union[str, None] = None,
        response_class: Union[type[BaseModel], None] = None,
        status_code: Union[int, None] = None,
        tags: Union[list[str], None] = None,
        summary: Union[str, None] = None,
        description: Union[str, None] = None,
        use_vpc: bool = True,
        skip_validation: bool = False,
    ) -> Decorator:
        return self.route(
            method=Methods.GET,
            path=path,
            name=name,
            authorizer=authorizer,
            response_class=response_class,
            status_code=status_code,
            tags=tags,
            summary=summary,
            description=description,
            use_vpc=use_vpc,
            skip_validation=skip_validation,
        )

    def post(
        self: Self,
        path: str,
        name: str,
        *,
        authorizer: Union[str, None] = None,
        response_class: Union[type[BaseModel], None] = None,
        status_code: Union[int, None] = None,
        tags: Union[list[str], None] = None,
        summary: Union[str, None] = None,
        description: Union[str, None] = None,
        use_vpc: bool = True,
        skip_validation: bool = False,
    ) -> Decorator:
        return self.route(
            method=Methods.POST,
            path=path,
            name=name,
            authorizer=authorizer,
            response_class=response_class,
            status_code=status_code,
            tags=tags,
            summary=summary,
            description=description,
            use_vpc=use_vpc,
            skip_validation=skip_validation,
        )

    def delete(
        self: Self,
        path: str,
        name: str,
        *,
        authorizer: Union[str, None] = None,
        response_class: Union[type[BaseModel], None] = None,
        status_code: Union[int, None] = None,
        tags: Union[list[str], None] = None,
        summary: Union[str, None] = None,
        description: Union[str, None] = None,
        use_vpc: bool = True,
        skip_validation: bool = False,
    ) -> Decorator:
        return self.route(
            method=Methods.DELETE,
            path=path,
            name=name,
            authorizer=authorizer,
            response_class=response_class,
            status_code=status_code,
            tags=tags,
            summary=summary,
            description=description,
            use_vpc=use_vpc,
            skip_validation=skip_validation,
        )

    def put(
        self: Self,
        path: str,
        name: str,
        *,
        authorizer: Union[str, None] = None,
        response_class: Union[type[BaseModel], None] = None,
        status_code: Union[int, None] = None,
        tags: Union[list[str], None] = None,
        summary: Union[str, None] = None,
        description: Union[str, None] = None,
        use_vpc: bool = True,
        skip_validation: bool = False,
    ) -> Decorator:
        return self.route(
            method=Methods.PUT,
            path=path,
            name=name,
            authorizer=authorizer,
            response_class=response_class,
            status_code=status_code,
            tags=tags,
            summary=summary,
            description=description,
            use_vpc=use_vpc,
            skip_validation=skip_validation,
        )

    def patch(
        self: Self,
        path: str,
        name: str,
        *,
        authorizer: Union[str, None] = None,
        response_class: Union[type[BaseModel], None] = None,
        status_code: Union[int, None] = None,
        tags: Union[list[str], None] = None,
        summary: Union[str, None] = None,
        description: Union[str, None] = None,
        use_vpc: bool = True,
        skip_validation: bool = False,
    ) -> Decorator:
        return self.route(
            method=Methods.PATCH,
            path=path,
            name=name,
            authorizer=authorizer,
            response_class=response_class,
            status_code=status_code,
            tags=tags,
            summary=summary,
            description=description,
            use_vpc=use_vpc,
            skip_validation=True,
        )


class SpyAPI(_SPY):
    def __init__(
        self: Self,
        *,
        config: ServerlessConfig,
        environment: Union[dict[str, Any], None] = None,
        title: Union[str, None] = None,
        version: Union[str, None] = None,
        prefix: Union[str, None] = None,
    ) -> None:
        super().__init__(prefix)

        self.title = title or "My API"
        self.version = version or "v0.0.1"
        self.config = config
        self.environment = environment


class SpyRouter(_SPY):
    def __init__(self: Self, prefix: Union[str, None] = None) -> None:
        super().__init__(prefix)



import pytest

from aws_spy import ServerlessConfig, SpyAPI, SpyRouter, __version__
from aws_spy.core.schemas import Methods


def test_app() -> None:
    assert __version__ == "0.1.8"


def test_app_conf(app: SpyAPI, config: ServerlessConfig) -> None:
    assert app.config == config


@pytest.mark.parametrize("prefix", ["api", "/api", "/somet_other_prefix"])
def test_prefix(config: ServerlessConfig, prefix: str) -> None:
    app = SpyAPI(config=config, prefix=prefix)
    expected_prefix = prefix if prefix.startswith("/") else f"/{prefix}"
    assert app.prefix == expected_prefix


@pytest.mark.parametrize("method", [method for method in Methods])
def test_register_router(config: ServerlessConfig, method: Methods) -> None:
    app = SpyAPI(config=config, prefix="/api")
    router = SpyRouter(prefix="/router")
    router_method = getattr(router, method.value)

    @router_method("path", "lambda")
    def handler() -> None:
        ...

    assert len(router.routes) == 1
    assert len(app.routes) == 0
    assert router.routes["/router/path"][method]

    app.register_router(router)
    assert len(app.routes) == 1
    assert app.routes["/api/router/path"][method]



import inspect
import os
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Literal, TypeVar, Union

from pydantic import BaseModel, Field, root_validator, validator
from typing_extensions import Self  # type: ignore

from aws_spy.core.exceptions import RouteDefinitionException
from aws_spy.core.schemas_utils import (
    ParamSchema,
    get_path_param_names,
    resolve_handler_args,
)

# pydantic_yaml won't be in the layer
try:
    from pydantic_yaml import YamlModel
except ImportError:  # pragma: no cover

    class YamlModel(BaseModel):  # type: ignore
        ...


LHReturnType = TypeVar("LHReturnType")
LH = Callable[..., LHReturnType]  # Lambda Handler
Decorator = Callable[[LH], LH]


class Methods(str, Enum):
    GET = "get"
    POST = "post"
    DELETE = "delete"
    PUT = "put"
    PATCH = "patch"


DEFAULT_STATUS_CODES = {
    Methods.GET: "200",
    Methods.POST: "201",
    Methods.DELETE: "204",
    Methods.PUT: "200",
    Methods.PATCH: "200",
}


class Functions(str, Enum):
    LAYER = "layer"
    OPENAPI = "openapi"
    SLS = "sls"


class SpyRoute(BaseModel):
    name: str
    path: str
    method: Methods
    handler: LH
    status_code: int
    tags: Union[list[str], None]
    summary: str = Field("API endpoint")
    description: Union[str, None]
    request_body_arg_name: Union[str, None]
    request_body: Union[type[BaseModel], None]
    response_class: Union[type[BaseModel], None]
    header_params: list[ParamSchema] = Field(default_factory=list)
    path_params: list[ParamSchema] = Field(default_factory=list)
    query_params: list[ParamSchema] = Field(default_factory=list)
    use_vpc: bool = Field(True)
    authorizer: Union[str, None] = Field(None)
    layers: list[str] = Field(default_factory=list)
    add_event: bool = Field(default=False)
    add_context: bool = Field(default=False)
    skip_validation: bool = Field(default=False)

    @root_validator(pre=True)
    def set_status_code(
        cls: type[Self],
        values: dict[str, Any],
    ) -> dict[str, Any]:
        if values.get("status_code") is None:
            values["status_code"] = DEFAULT_STATUS_CODES[values["method"]]

        return values

    @validator("summary", pre=True)
    def set_summary(cls: type[Self], summary: Union[str, None]) -> str:
        return summary or "API endpoint"

    @root_validator
    def validate_handler_params(
        cls: type[Self],
        values: dict[str, Any],
    ) -> dict[str, Any]:
        method: Methods = values["method"]
        path: str = values["path"]
        handler: LH = values["handler"]

        path_params = get_path_param_names(path)
        handler_args = resolve_handler_args(handler)

        args = inspect.signature(handler).parameters
        args_count = len(args)
        # exclude lambdas event and context from count
        if "event" in args:
            values["add_event"] = True
            args_count -= 1
        if "context" in args:
            values["add_context"] = True
            args_count -= 1
        if handler_args.count != args_count and not values["skip_validation"]:
            raise RouteDefinitionException(
                f'Unrecognized params for {method.upper()} method on "{path}" path!'
            )

        for path_param in path_params:
            if path_param not in handler_args.path.keys():
                raise RouteDefinitionException(
                    f"You did not specify {path_param} in your handler arguments!"
                )

        for path_arg in handler_args.path.keys():
            if path_arg not in path_params:
                raise RouteDefinitionException(
                    f"Your {path_arg} path parameter is missing in "
                    f'{method.upper()} method on "{path}" path!'
                )

        if method in (Methods.GET, Methods.DELETE) and handler_args.request_body:
            raise RouteDefinitionException(
                f'{method.upper()} method on "{path}" cannot have request body!'
            )

        if handler_args.request_body:
            values["request_body"] = handler_args.request_body
            values["request_body_arg_name"] = handler_args.request_body_arg_name

        for attr_name in ("path", "query", "header"):
            values[f"{attr_name}_params"] += [
                param for _, param in getattr(handler_args, attr_name).items()
            ]

        return values


class _CloudFormationRef(BaseModel):
    stack_name: str
    export_name: str

    def __str__(self: Self) -> str:
        return f"${{cf:{self.stack_name}-${{opt:stage}}.{self.export_name}}}"


class CloudFormationRef(_CloudFormationRef):
    def __new__(cls: type[Self], stack_name: str, export_name: str) -> str:  # type: ignore
        instance = _CloudFormationRef(stack_name=stack_name, export_name=export_name)
        return str(instance)


class _JSONFileRef(BaseModel):
    file_path: str
    field: str

    def __str__(self: Self) -> str:
        return f"${{file({self.file_path}):{self.field}}}"


class JSONFileRef(_JSONFileRef):
    def __new__(cls: type[Self], file_path: str, field: str) -> str:  # type: ignore
        instance = _JSONFileRef(file_path=file_path, field=field)
        return str(instance)


def build_cognito_issue_url(
    user_pool_id: Union[str, CloudFormationRef, JSONFileRef]
) -> str:
    return f"https://cognito-idp.${{region}}.amazonaws.com/{user_pool_id}"


class Authorizer(BaseModel):
    type: Literal["jwt"] = "jwt"
    identitySource: str = "$request.header.Authorization"
    issuerUrl: str
    audience: list[Union[str, CloudFormationRef, JSONFileRef]]


class CORS(BaseModel):
    allowedHeaders: list[str]
    exposedResponseHeaders: list[str]
    allowedMethods: list[str]
    allowedOrigins: list[str]


class HTTPApi(BaseModel):
    authorizers: dict[str, Authorizer]
    cors: CORS


class VPC(BaseModel):
    securityGroupIds: list[Union[str, CloudFormationRef, JSONFileRef]]
    subnetIds: list[Union[str, CloudFormationRef, JSONFileRef]]


class Function(BaseModel):
    handler: str
    module: str
    events: list[dict[str, Any]]
    layers: list[Union[str, CloudFormationRef, JSONFileRef]]
    environment: Union[dict[str, Any], None] = Field(None)

    @staticmethod
    def generate_rel_path_for_function(route: SpyRoute) -> str:
        return os.path.relpath(
            Path(route.handler.__code__.co_filename), Path().resolve()
        )

    @classmethod
    def from_route(
        cls: type[Self], *, route: SpyRoute, path: str, method: Methods
    ) -> Self:
        rel_path = cls.generate_rel_path_for_function(route)
        http_api_event: dict[str, Any] = {"path": path, "method": method.upper()}
        if route.authorizer:
            http_api_event["authorizer"] = {"name": route.authorizer}

        return cls(
            handler=(
                f'{rel_path.split(os.sep)[-1].replace(".py", "")}'
                f".{route.handler.__name__}"
            ),
            module="/".join(rel_path.split(os.sep)[:-1]),
            events=[{"httpApi": http_api_event}],
            layers=list(
                set(
                    [
                        CloudFormationRef(
                            stack_name="spy-layer",
                            export_name="ServerlesspyLayerExport",
                        )
                    ]
                    + route.layers  # type: ignore
                )
            ),
        )


class Provider(BaseModel):
    name: str = "aws"
    runtime: str = "python3.9"
    region: str
    role: Union[str, CloudFormationRef, JSONFileRef]
    httpApi: HTTPApi
    vpc: Union[VPC, None]


class ServerlessConfig(YamlModel):
    service: str
    custom: Union[dict[str, Any], None] = Field(None)
    plugins: list[str]
    configValidationMode: str = "error"
    provider: Provider
    package: dict[str, bool] = Field({"individually": True})
    functions: Union[dict[str, Function], None] = Field(None)

    @validator("plugins", pre=True)
    def set_default_plugins(
        cls: type[Self], value: Union[list[str], None]
    ) -> list[str]:
        value = value if value is not None else []
        value += [
            "serverless-python-requirements",
            "serverless-plugin-common-excludes",
            "serverless-plugin-include-dependencies",
        ]
        return list(set(value))



import json
import os
import sys
from argparse import ArgumentParser
from distutils.dir_util import copy_tree
from distutils.file_util import copy_file
from functools import wraps
from pathlib import Path
from typing import Callable, get_args

from aws_spy import SpyAPI
from aws_spy.core import logger
from aws_spy.core.exceptions import RouteDefinitionException
from aws_spy.core.schemas import Function, Functions
from aws_spy.core.types import is_type_required
from aws_spy.helpers.documentation import get_openapi
from aws_spy.helpers.exceptions import (
    PythonEnvironmentException,
    WrongArgumentException,
)
from aws_spy.helpers.utils import LoadAppFromStringError, load_app_from_string


def unpack_args(function: Callable[..., None]) -> Callable[..., None]:
    @wraps(function)
    def wrapper(*_, **kwargs: ArgumentParser) -> None:
        parser = kwargs["parser"]
        function_kwargs = {}
        for argument_name, argument_type_hint in function.__annotations__.items():
            if argument_name == "return":
                continue

            if argument_name == "app":
                parser.add_argument("app", type=str)
                args, __ = parser.parse_known_args()
                try:
                    function_kwargs["app"] = load_app_from_string(args.app)
                    continue
                except LoadAppFromStringError as e:
                    logger.error(e)
                    return

            is_required = is_type_required(argument_type_hint)
            argument_type = (
                argument_type_hint if is_required else get_args(argument_type_hint)[0]
            )

            parser.add_argument(
                f"-{argument_name[0]}",
                f"--{argument_name}",
                type=argument_type,
                required=is_required,
            )
        function_kwargs.update(
            {
                name: value
                for name, value in parser.parse_args().__dict__.items()
                if name not in ("function", "app")
            }
        )
        return function(**function_kwargs)

    return wrapper


@unpack_args
def deploy_layer(stage: str, region: str) -> None:
    site_packages = None
    for path in sys.path:
        if path.endswith("site-packages"):
            site_packages = path

    if site_packages is None:
        raise PythonEnvironmentException("Couldn't find site-packages folder.")
    spy_package_name = "aws_spy"

    serverlesspy_path = Path(os.path.join(site_packages, spy_package_name))
    layer_path = Path(os.path.join(serverlesspy_path, "layer", "spy", "python"))
    serverless_layer_path = Path(os.path.join(layer_path, spy_package_name))
    serverless_layer_path.mkdir(parents=True, exist_ok=True)

    for package in ("pydantic",):
        package_path = Path(os.path.join(site_packages, package))
        if not package_path.is_dir():
            raise PythonEnvironmentException(
                f'Could not find "{package}" package. '
                'Make sure it is installed in your current environment.'
            )
        copy_tree(str(package_path), os.path.join(layer_path, package))

    copy_file(
        os.path.join(site_packages, "typing_extensions.py"),
        os.path.join(layer_path, "typing_extensions.py"),
    )

    copy_tree(
        os.path.join(serverlesspy_path, "core"),
        os.path.join(serverless_layer_path, "core"),
    )
    for file_to_copy in ("__init__.py", "main.py", "responses.py"):
        copy_file(
            os.path.join(serverlesspy_path, file_to_copy),
            os.path.join(serverless_layer_path, file_to_copy),
        )
    current_working_dir = os.getcwd()
    os.chdir(os.path.join(serverlesspy_path, "layer"))
    os.system(f"serverless deploy -s {stage} -c spy-layer.yml --region {region}")
    os.chdir(current_working_dir)


@unpack_args
def generate_openapi(app: SpyAPI, path: str) -> None:
    open_api = get_openapi(app.title, app.version, app.routes)
    with open(f"./{path}", "w") as file:
        json.dump(open_api, file)


@unpack_args
def generate_serverless_file(app: SpyAPI, path: str) -> None:
    if not path.endswith(".yml"):
        raise WrongArgumentException("File is not YAML file.")

    functions: dict[str, Function] = {}
    for route_path, route_dict in app.routes.items():
        for method, route in route_dict.items():
            if (
                route.authorizer is not None
                and route.authorizer
                not in app.config.provider.httpApi.authorizers.keys()
            ):
                raise RouteDefinitionException(
                    f"Authorizer {route.authorizer} not defined"
                )

            functions[route.name] = Function.from_route(
                route=route, method=method, path=route_path
            )

    app.config.functions = functions

    with open(path, "w") as file:
        file.write(app.config.yaml(exclude_none=True))


FUNCTIONS_DEFINITIONS: dict[str, Callable[..., None]] = {
    "layer": deploy_layer,
    "openapi": generate_openapi,
    "sls": generate_serverless_file,
}


def cli_handler() -> None:
    parser = ArgumentParser()
    parser.add_argument("function", type=Functions)
    args, _ = parser.parse_known_args()

    return FUNCTIONS_DEFINITIONS[args.function](parser=parser)



import inspect
import re
from enum import Enum
from typing import Any, Callable, Set, TypeVar, Union

from pydantic import BaseModel
from typing_extensions import Self  # type: ignore

from aws_spy.core import types
from aws_spy.core.exceptions import RouteDefinitionException
from aws_spy.core.params import Param, ParamType

LH = TypeVar("LH", bound=Callable[..., Any])


class ParamSchema(BaseModel):
    name: str
    arg_name: str
    in_: Param
    annotation: type
    is_required: bool
    enum: Union[list[str], None]

    class Config:
        arbitrary_types_allowed = True


class HandlerArgs(BaseModel):
    query: dict[str, ParamSchema]
    path: dict[str, ParamSchema]
    header: dict[str, ParamSchema]
    request_body: Union[type[BaseModel], None]
    request_body_arg_name: Union[str, None]

    @property
    def count(self: Self) -> int:
        return (
            len(self.query)
            + len(self.path)
            + len(self.header)
            + (1 if self.request_body is not None else 0)
        )


def _is_request_body(arg) -> bool:
    return arg.default is inspect.Parameter.empty and issubclass(
        arg.annotation, BaseModel
    )


def _is_param(arg) -> bool:
    return isinstance(arg.default, Param)


def resolve_handler_args(handler: LH) -> HandlerArgs:
    params: dict[ParamType, dict[str, ParamSchema]] = {
        ParamType.HEADER: {},
        ParamType.PATH: {},
        ParamType.QUERY: {},
    }
    request_body = None
    request_body_arg_name = None
    for arg_name, arg_value in inspect.signature(handler).parameters.items():
        if _is_request_body(arg_value) and request_body is None:
            request_body = arg_value.annotation
            request_body_arg_name = arg_name
            continue
        if _is_param(arg_value):
            param: Param = arg_value.default
            param_name = param.name if param.name is not None else arg_name
            if param_name in params[param.in_].keys():
                raise RouteDefinitionException(
                    f"{handler.__name__} expects two same "
                    f'{param.in_} params: "{param_name}"!'
                )

            enum = None
            is_required = types.is_type_required(arg_value.annotation)
            annotation = (
                arg_value.annotation
                if is_required
                else types.get_type_from_optional(arg_value.annotation)
            )
            if issubclass(annotation, Enum):
                enum = [e.value for e in annotation]

            params[param.in_][param_name] = ParamSchema(
                name=param_name,
                in_=arg_value.default,
                arg_name=arg_name,
                annotation=annotation,
                enum=enum,
                is_required=is_required,
            )

    return HandlerArgs(
        query=params[ParamType.QUERY],
        path=params[ParamType.PATH],
        header=params[ParamType.HEADER],
        request_body=request_body,
        request_body_arg_name=request_body_arg_name,
    )


def get_path_param_names(path: str) -> Set[str]:
    return set(re.findall("{(.*?)}", path))
